/* keyboard.c - 6502 microcomputer emulator
 *
 * Copyright (C) 2017, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * -----
 *
 * Credit:
 *
 * The polled keyboard code in this file was written after taking a look
 * at Tim Baldwin's excellent Java-based UK101 Simulator, and was heavily
 * influenced by that. Tim Baldwin's project is BSD licensed, and can be
 * found here: http://uk101.sourceforge.net/
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "xlib/xutil.h"
#include "xlib/xmem.h"
#include "xlib/xlist.h"

#include "config.h"
#include "memory.h"
#include "machine.h"
#include "keyboard.h"

#ifdef ENABLE_SDL
#include <SDL2/SDL.h>
#endif

#ifdef ENABLE_NCURSES
#include <ncurses.h>
#endif


#ifdef DEBUG
#define DEBUG_KEYBOARD   0
#endif


#define KEY_BUFFER_SIZE  10 /* if ncurses */

#define MATRIX_LEN       8

#define IS_POLLED_OSI(kb_type) (kb_type == KEYBOARD_TYPE_OSI542     \
                             || kb_type == KEYBOARD_TYPE_OSI600     \
                             || kb_type == KEYBOARD_TYPE_UK101      \
                             || kb_type == KEYBOARD_TYPE_UK101_US)


struct _Keyboard {

	/* Memory */

	MemoryType     type;

	int            base;
	int            size;
	xbool_t        read_only;

	MemoryFreeData_f  free_data;
	MemoryStoreByte_f store_byte;
	MemoryReadByte_f  read_byte;

	Byte           *data;

	/* Keyboard */

	KeyboardType   keyboard_type;
	xlist_t        *keys;
	Byte           matrix[MATRIX_LEN];

	xbool_t        shift_lock;

	xbool_t        reset;
	xbool_t        stepping;
	xbool_t        step_now;
	xbool_t        fullscreen_toggle;

#ifdef ENABLE_NCURSES
	xlist_t        *key_buffer;
#endif

};


typedef struct {

	int id;
	int row;
	int col;

} PolledKey;


extern Machine *machine;


static void       init_polled_osi_keyboard (Keyboard *k);
static void       key_new (Keyboard *k, int id1, int id2, int row, int col);
static PolledKey* key_get (Keyboard *k, int id);


static void
keyboard_store_byte (Keyboard *k, Word offset, Byte b)
{
	if (IS_POLLED_OSI (k->keyboard_type)) {
		mem_store_byte ((Memory*)k, 0, b);
	}
}


static Byte
keyboard_read_byte (Keyboard *k, Word offset)
{
	Byte b;


	if (IS_POLLED_OSI (k->keyboard_type)) {

		Byte poll_row;
		int i;

		b = 0xFF;

		poll_row = k->data[0];

		for (i = 0; i < MATRIX_LEN; i++) {
			if ((poll_row & 1) == 0) {
				b &= k->matrix[i];
			}
			poll_row = poll_row >> 1;
		}

	}
	else if (k->keyboard_type == KEYBOARD_TYPE_NCURSES) {

#ifdef ENABLE_NCURSES

		int *ch;

		if (k->key_buffer != NULL) {
			ch = xlist_pop (&k->key_buffer);
			b = (Byte)*ch;
		}
		else {
			b = 0;
		}

#else
		b = 0;
#endif

	}
	else if (k->keyboard_type == KEYBOARD_TYPE_STDIO) {

		b = fgetc (stdin);

	}
	else {
		b = 0;
	}

	return b;
}


Keyboard*
keyboard_new (Profile *p)
{
	Keyboard *k;
	Word n_bytes;

	k = xnew0 (Keyboard, 1);
	k->keyboard_type = p->keyboard_type;

	n_bytes = 1;

	mem_init ((Memory*)k, p->keyboard_base, n_bytes, FALSE, MEM_TYPE_KEYBOARD);

	mem_set_store_func ((Memory*)k, (MemoryStoreByte_f)keyboard_store_byte);
	mem_set_read_func ((Memory*)k, (MemoryReadByte_f)keyboard_read_byte);
	/* mem_set_free_func ((Memory*)k, (MemoryStoreByte_f)keyboard_free_data); */

	if (IS_POLLED_OSI(k->keyboard_type)) {
		init_polled_osi_keyboard (k);
	}

	return k;
}


void
keyboard_free (Keyboard *k)
{
	xlist_t *this;

	if (k == NULL) {
		return;
	}

	if (k->keys != NULL) {
		for (this = k->keys; this != NULL; this = this->next) {
			xfree (this->data);
		}
		xlist_free (&k->keys);
	}

#ifdef ENABLE_NCURSES
	if (k->key_buffer != NULL) {
		for (this = k->key_buffer; this != NULL; this = this->next) {
			xfree (this->data);
		}
		xlist_free (&k->key_buffer);
	}
#endif

	k->free_data ((Memory*)k);
	xfree (k);
}


static PolledKey*
key_get (Keyboard *k, int id)
{
	xlist_t *this;

	for (this = k->keys; this != NULL; this = this->next) {
		if (((PolledKey*)this->data)->id == id) {
			return this->data;
		}
	}
	return NULL;
}


void
keyboard_key_press (Keyboard *k, int id)
{
	PolledKey *key;

	if ((key = key_get (k, id)) != NULL) {
		k->matrix[key->row] &= ~key->col;
#if DEBUG_KEYBOARD
		fprintf (stderr, "pressed:  [%c] (%03d); ",
		         isprint (id) ? id : ' ', id);
		fprintf (stderr, "row = %d, col = %d, matrix[%d] = %d\n",
		         key->row, key->col, key->row, k->matrix[key->row]);
#endif
	}
}


void
keyboard_key_release (Keyboard *k, int id)
{
	PolledKey *key;

	if ((key = key_get (k, id)) != NULL) {
		k->matrix[key->row] |= key->col;
#if DEBUG_KEYBOARD
		fprintf (stderr, "released: [%c] (%03d); ",
		         isprint (id) ? id : ' ', id);
		fprintf (stderr, "row = %d, col = %d, matrix[%d] = %d\n",
		         key->row, key->col, key->row, k->matrix[key->row]);
#endif
	}
}


static void
key_new (Keyboard *k, int id1, int id2, int row, int col)
{
	PolledKey *key;

	if (id1 != 0) {

		key = xnew (PolledKey, 1);

		key->id  = id1;
		key->row = row;
		key->col = (Byte)(1 << col);

		xlist_append (&k->keys, key);

	}

	if (id2 != 0) {

		key = xnew (PolledKey, 1);

		key->id  = id2;
		key->row = row;
		key->col = (Byte)(1 << col);

		xlist_append (&k->keys, key);

	}

#if DEBUG_KEYBOARD
	fprintf (stderr, "add key [%c][%c] row = %d, col = %d\n",
	    isprint (id1) ? id1 : ' ',
	    isprint (id2) ? id2 : ' ',
	    key->row, key->col);
#endif

}


xbool_t
keyboard_get_reset (Keyboard *k)
{
	return k->reset;
}


void
keyboard_set_reset (Keyboard *k, xbool_t reset)
{
	k->reset = reset;
}


xbool_t
keyboard_get_stepping (Keyboard *k)
{
	return k->stepping;
}


void
keyboard_set_stepping (Keyboard *k, xbool_t stepping)
{
	k->stepping = stepping;
}


xbool_t
keyboard_get_step_now (Keyboard *k)
{
	return k->step_now;
}


void
keyboard_set_step_now (Keyboard *k, xbool_t step_now)
{
	k->step_now = step_now;
}


void
keyboard_set_fullscreen_toggle (Keyboard *k, xbool_t toggle)
{
	k->fullscreen_toggle = toggle;
}


xbool_t
keyboard_get_fullscreen_toggle (Keyboard *k)
{
	return k->fullscreen_toggle;
}


static void
init_polled_osi_keyboard (Keyboard *k)
{
	Byte b;
	int i;

	b = (k->keyboard_type == KEYBOARD_TYPE_OSI542 ? 0x00 : 0xFF);

	for (i = 0; i < MATRIX_LEN; i++) {
		k->matrix[i] = b;
	}

	key_new (k, '1', '!',  7, 7);
	key_new (k, '2', '\"', 7, 6);
	key_new (k, '3', '#',  7, 5);
	key_new (k, '4', '$',  7, 4);
	key_new (k, '5', '%',  7, 3);
	key_new (k, '6', '&',  7, 2);
	key_new (k, '7', '\'', 7, 1);
	key_new (k, '8', '(',  6, 7);
	key_new (k, '9', ')',  6, 6);
	key_new (k, '0', '@',  6, 5);
	key_new (k, ':', '*',  6, 4);
	key_new (k, '-', '=',  6, 3);
	key_new (k, EMU_KEY_RUB_OUT, 0, 6, 2);
	key_new (k, '.', '>',  5, 7);
	key_new (k, 'L', 'l',  5, 6);  
	key_new (k, 'O', 'o',  5, 5);
	key_new (k, EMU_KEY_RETURN, 0, 5, 3);
	key_new (k, 'W', 'w',  4, 7);
	key_new (k, 'E', 'e',  4, 6);
	key_new (k, 'R', 'r',  4, 5);
	key_new (k, 'T', 't',  4, 4);
	key_new (k, 'Y', 'y',  4, 3);
	key_new (k, 'U', 'u',  4, 2);
	key_new (k, 'I', 'i',  4, 1);
	key_new (k, 'S', 's',  3, 7);
	key_new (k, 'D', 'd',  3, 6);
	key_new (k, 'F', 'f',  3, 5);
	key_new (k, 'G', 'g',  3, 4);
	key_new (k, 'H', 'h',  3, 3);
	key_new (k, 'J', 'j',  3, 2);
	key_new (k, 'K', 'k',  3, 1);  
	key_new (k, 'X', 'x',  2, 7);
	key_new (k, 'C', 'c',  2, 6);
	key_new (k, 'V', 'v',  2, 5);
	key_new (k, 'B', 'b',  2, 4);
	key_new (k, 'N', 'n',  2, 3);
	key_new (k, 'M', 'm',  2, 2);  
	key_new (k, ',', '<',  2, 1);
	key_new (k, 'Q', 'q',  1, 7);
	key_new (k, 'A', 'a',  1, 6);
	key_new (k, 'Z', 'z',  1, 5);
	key_new (k, EMU_KEY_SPACE, ' ', 1, 4);
	key_new (k, '/', '?',  1, 3);
	key_new (k, ';', '+',  1, 2);
	key_new (k, 'P', 'p',  1, 1);  
	key_new (k, EMU_KEY_CTRL,  0, 0, 6);
	key_new (k, EMU_KEY_SHIFT_LT, 0, 0, 2);
	key_new (k, EMU_KEY_SHIFT_RT, 0, 0, 1);
	key_new (k, EMU_KEY_SHIFT_LOCK, 0, 0, 0);

	key_new (k, '\\', 0, 5, 6);
	key_new (k, '_', 0, 5, 5);
	key_new (k, '[', 0, 3, 1);
	key_new (k, ']', 0, 2, 2);

	if (k->keyboard_type == KEYBOARD_TYPE_OSI600) {
		key_new (k, EMU_KEY_LINE_FEED, 0, 5, 4);
		key_new (k, EMU_KEY_ESC, 0, 0, 5);
		key_new (k, EMU_KEY_REPEAT, 0, 0, 7);
		key_new (k, '^', 0, 2, 3);
	}
	else if (k->keyboard_type == KEYBOARD_TYPE_UK101_US) {
		key_new (k, EMU_KEY_UP_ARROW, '^', 5, 4);
	}

	keyboard_key_press (k, EMU_KEY_SHIFT_LOCK);
	k->shift_lock = TRUE;

}


#ifdef ENABLE_SDL
void
keyboard_event_sdl (Keyboard *keyboard, int key, xbool_t down)
{
	/* Remap SDL key codes to our OSI-style keyboard codes */

	switch (key) {
	case SDLK_RETURN:      key = EMU_KEY_RETURN; break;
	case SDLK_LSHIFT:      key = EMU_KEY_SHIFT_LT; break;
	case SDLK_RSHIFT:      key = EMU_KEY_SHIFT_RT; break;
	case SDLK_BACKSPACE:   key = EMU_KEY_RUB_OUT; break;
	case SDLK_INSERT:      key = EMU_KEY_LINE_FEED; break;
	case SDLK_DELETE:      key = EMU_KEY_REPEAT; break;
	case SDLK_CAPSLOCK:    key = EMU_KEY_SHIFT_LOCK; break;
	case SDLK_LCTRL:
	case SDLK_RCTRL:       key = EMU_KEY_CTRL; break;
	case SDLK_ESCAPE:      key = EMU_KEY_ESC; break;
	case SDLK_PAUSE:       key = EMU_KEY_BREAK; break;

	case SDLK_KP_MINUS:    key = EMU_KEY_BREAK; break;

	case SDLK_MINUS:       key = ':'; break;
	case SDLK_UNDERSCORE:  key = '*'; break;

	/* A couple of keys are used to control the emulator: */

	case SDLK_KP_MULTIPLY: key = EMU_KEY_STEP_TOGGLE; break;
	case SDLK_KP_PLUS:     key = EMU_KEY_STEP; break;
	case SDLK_F11:         key = EMU_KEY_FULLSCREEN_TOGGLE; break;
	default: break;
	}

	if (!down && key == EMU_KEY_SHIFT_LOCK) {
		if (keyboard->shift_lock) {
			keyboard_key_release (keyboard, key);
		}
		else {
			keyboard_key_press (keyboard, key);
		}
		keyboard->shift_lock = ! keyboard->shift_lock;
		return;
	}

	if (down) {
		keyboard_key_press (keyboard, key);
		if (key == EMU_KEY_BREAK) {
			keyboard_set_reset (keyboard, TRUE);
		}
		else if (key == EMU_KEY_STEP_TOGGLE) {
			keyboard_set_stepping (keyboard, TRUE);
		}
		else if (key == EMU_KEY_STEP) {
			keyboard_set_step_now (keyboard, TRUE);
		}
		else if (key == EMU_KEY_FULLSCREEN_TOGGLE) {
			keyboard_set_fullscreen_toggle (keyboard, TRUE);
		}
	}
	else {
		keyboard_key_release (keyboard, key);
	}
}
#endif


#ifdef ENABLE_NCURSES
void
keyboard_event_ncurses (Keyboard *keyboard, int key)
{
	int *ch;

	if (keyboard->key_buffer
	    && xlist_length (keyboard->key_buffer) <= KEY_BUFFER_SIZE) {

		ch = xnew (int, sizeof (int));
		*ch = key;

		xlist_push (&keyboard->key_buffer, ch);

	}

}
#endif


