/* rom.c - 6502 microcomputer emulator
 *
 * Copyright (C) 2017, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <assert.h>

#include "xlib/xutil.h"
#include "xlib/xmem.h"
#include "xlib/xstr.h"
#include "xlib/xfile.h"

#include "config.h"
#include "support.h"
#include "memory.h"
#include "rom.h"


struct _ROM {

	/* Memory */

	MemoryType     type;

	int            base;
	int            size;
	xbool_t        read_only;

	MemoryFreeData_f  free_data;
	MemoryStoreByte_f store_byte;
	MemoryReadByte_f  read_byte;

	Byte           *data;

	/* ROM */

	/* No special features */

};


#ifdef DEBUG
extern xbool_t mem_quiet;
#endif


static char* parse_rom_info (const char *info, int *base, int *size);


ROM*
rom_new (Profile *p, const char *file_info)
{
	char *fname, *fullpath;
	int base, size, n;
	ROM *r = NULL;

	fname = parse_rom_info (file_info, &base, &size);

	if (fname == NULL || size == -1) {
		fprintf (stderr, "%s: Trouble parsing ROM details (%s) from '%s'\n",
		         p->progname, file_info, p->basename);
	}
	else {

		fullpath = path_to_file (p->dirname, p->basename, fname);
#if 0
		fullpath = xstr_join (DIR_SEPARATOR_S,
		                      p->dirname, p->basename, fname, NULL);
#endif

		r = xnew0 (ROM, 1);
		mem_init ((Memory*)r, base, size, FALSE, MEM_TYPE_ROM);

		n = load_hex (p->progname, fullpath, (Memory*)r, 0);

		xfree (fullpath);

		if (n < 0) {
			xfree (r);
			r = NULL;
		}
		else {

			/* Now that the ROM is loaded, mark it read-only */

			r->read_only = TRUE;
		}
#ifdef DEBUG
		if (r != NULL && !mem_quiet) {
			fprintf (stderr, " * Loaded %d bytes from %s\n", n, fname);
		}
#endif
	}

	xfree (fname);

	return r;
}


void
rom_free (ROM *r)
{
	if (r == NULL) {
		return;
	}

	r->free_data ((Memory*)r);
	xfree (r);
}


static char*
parse_rom_info (const char *info, int *base, int *size)
{
	char **split;
	char *fname = NULL;
	int n;

	assert (info != NULL);

	*base = -1;
	*size = -1;

	split = xstr_split (info, ", ");
	for (n = 0; split[n] != NULL && *split[n] != '\0'; n++);

	if (n > 0) {
		fname = xstr_dup (split[0]);
	}

	if (n > 1) {
		*size = parse_ascii_integer (split[1]);
	}

	if (n > 2) {
		*base = parse_ascii_integer (split[2]);
	}

	if (n > 3) {
		fprintf (stderr, "Error in profile, unable to load ROM '%s'\n", fname);
		xfree (fname);
		fname = NULL;
	}

	xstr_freev (split);

	return fname;
}

