/* xstr.c - string utilities
 *
 * Copyright (C) 2005, 2011, 2017, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * gcc -D TEST xmem.o xstr.c -o xstr-test
 *
 */
 

#define _GNU_SOURCE /* for vasprintf */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <err.h>
#include <limits.h>

#ifndef XUTIL_H
#include "xutil.h"
#endif

#ifndef XMEM_H
#include "xmem.h"
#endif

#ifndef XSTR_H
#include "xstr.h"
#endif



/* Return a new string of length len, filled with char c and
   terminated with '\0'.  */

char*
xstr_nfill (size_t len, char c)
{
	char *s;

	s = xmalloc (sizeof(char) * (len + 1));

	s[len] = '\0';
	while (len > 0) {
		len--;
		s[len] = c;
	}

	return s;
}


/* Return a '\0' terminated copy of the first n chars of str.
   If str is NULL, return NULL.  */

char*
xstr_ndup (const char *str, size_t n)
{
	char *s;

	if (str == NULL) {
		return NULL;
	}

	s = xmalloc (sizeof(char) * (n + 1));
	memcpy (s, str, sizeof(char) * n);
	s[n] = '\0';

	return s;
}


/* Return a copy of str.
   If str is NULL, return NULL.  */

char*
xstr_dup (const char *str)
{
	char *s, *p;

	if (str == NULL) {
		return NULL;
	}

	s = p = xmalloc (sizeof(char) * (strlen (str) + 1));
	while ((*p++ = *str++) != '\0');

	return s;
}


/* Return a copy of str, converted to upper-case.
   If str is NULL, return NULL.  */

char*
xstr_dup_toupper (const char *str)
{
	char *s, *p;

	if (str == NULL) {
		return NULL;
	}

	s = p = xmalloc (sizeof(char) * (strlen (str) + 1));
	while ((*p++ = toupper (*str++)) != '\0');

	return s;
}


/* Return a copy of str, converted to lower-case.
   If str is NULL, return NULL.  */

char*
xstr_dup_tolower (const char *str)
{
	char *s, *p;

	if (str == NULL) {
		return NULL;
	}

	s = p = xmalloc (sizeof(char) * (strlen (str) + 1));
	while ((*p++ = tolower (*str++)) != '\0');

	return s;
}


/* Return TRUE if str begins with prefix, otherwise FALSE.  */

xbool_t
xstr_has_prefix (const char *str, const char *prefix)
{
	while (*str != '\0' && *prefix != '\0') {
		if (*str != *prefix) {
			break;
		}
		str++;
		prefix++;
	}
	return *prefix == '\0';
}


/* Return TRUE if str ends with suffix, otherwise FALSE.  */

xbool_t
xstr_has_suffix (const char *str, const char *suffix)
{
	int len_str, len_suffix;
	const char *p;

	len_str = strlen (str);
	len_suffix = strlen (suffix);

	if (len_suffix > len_str) {
		return FALSE;
	}

	p = str + len_str - len_suffix;
	while (*p != '\0' && *p == *suffix) {
		p++;
		suffix++;
	}

	return *p == '\0';
}


/* Remove leading spaces from str by shifting characters forward.  */

char*
xstr_chug (char *str)
{
	char *start = str;

	while (isspace (*start)) {
		start++;
	}

	if (start != str) {
		int len;
		len = strlen (start);
		memmove (str, start, len);
		str[len] = '\0';
	}

	return str;
}


/* Replace the first trailing space character in str with '\0'.  */

char*
xstr_chomp (char *str)
{
	char *end;

	end = str;
	while (*end != '\0') {
		end++;
	}

	if (end == str) {
		return str;
	}

	while (end > str && isspace (*(--end)));
	*(++end) = '\0';

	return str;
}


/* Remove leading and trailing space characters from str.  */

char*
xstr_strip (char *str)
{
	return xstr_chomp (xstr_chug (str));
}


/* Return a copy of str with max occurrences of
   substring f replaced by string r.
   If max is 0, all occurrences are replaced.  */

char*
xstr_replace (const char *str, const char *f, const char *r, int max)
{
	const char *start, *point;
	char *tmp, *new;
	size_t len, flen, rlen;
	int count;


	if ((point = strstr (str, f)) == NULL) {
		return xstr_dup (str);
	}

	rlen = strlen (r);
	flen = strlen (f);
	start = str;
	count = 0;

	tmp = new = xmalloc (sizeof(char) * (strlen (str) * rlen + 1));

	do {

		len = point - start;
		strncpy (tmp, start, len);
		tmp = tmp + len;

		strncpy (tmp, r, rlen);
		tmp = tmp + rlen;

		start = point + flen;

		if (max && ++count >= max) {
			break;
		}

		/* find next */
		point = strstr (start, f);

	} while (point != NULL);

	strcpy (tmp, start);

	tmp = xstr_dup (new);
	xfree (new);

	return tmp;
}


/* Concatenate arbitrarily many strings.
   Final argument must be NULL.
   (From glibc-doc/html/libc_5.html#SEC74)  */

char*
xstr_concat (const char *str, ...)
{
	va_list ap;
	size_t len = 1;
	const char *s;
	char *result;

	va_start (ap, str);

	for (s = str; s != NULL; s = va_arg (ap, const char*)) {
		len = len + strlen (s);
	}

	va_end (ap);

	result = xnew (char, len);
	result[0] = '\0';

	va_start (ap, str);

	for (s = str; s != NULL; s = va_arg (ap, const char*)) {
		strcat (result, s);
	}

	va_end (ap);

	return result;
}


/* Concatenate an array of strings.
   The final item in the array must be NULL.  */

char*
xstr_concatv (char **array)
{
	size_t len = 1;
	char **s, *result = NULL;

	s = array;
	while (*s != NULL) {
		len = len + strlen (*s);
		s++;
	}

	result = xnew (char, len);
	result[0] = '\0';

	while (*array != NULL) {
		strcat (result, *array);
		array++;
	}

	return result;
}


/* Concatenate arbitrarily many strings, placing
   a separator between each.
   The final argument must be NULL.  */

char*
xstr_join (const char *separator, ...)
{
	va_list ap;
	size_t len = 1;
	const char *s;
	char *result;
	int parts = 0;

	va_start (ap, separator);

	while ((s = va_arg (ap, const char*)) != NULL) {
		len = len + strlen (s);
		parts++;
	}
	if (parts == 0) {
		return NULL;
	}

	parts--;
	len = len + parts * strlen (separator);

	va_end (ap);

	result = xnew (char, len);
	result[0] = '\0';

	va_start (ap, separator);

	while ((s = va_arg (ap, const char*)) != NULL) {
		strcat (result, s);
		if (parts > 0) {
			strcat (result, separator);
		}
		parts--;
	}
	va_end (ap);

	return result;
}


/* Concatenate an array of strings, placing
   a separator between each.
   The final item in the array must be NULL.  */

char*
xstr_joinv (const char *separator, char **array)
{
	size_t len = 1;
	int parts = 0;
	char **s, *result = NULL;

	s = array;
	while (*s != NULL) {
		len = len + strlen (*s);
		parts++;
		s++;
	}

	parts--;
	len = len + parts * strlen (separator);

	result = xnew (char, len);
	result[0] = '\0';

	while (*array != NULL) {
		strcat (result, *array);
		if (parts > 0) {
			strcat (result, separator);
		}
		parts--;
		array++;
	}

	return result;
}


/* Split str into an array of strings and return that array.
   str is split wherever a character from delim is met.  */

char**
xstr_split (const char *str, const char *delim)
{
	char **array, *dup, *tok;
	size_t size, n = 1;

	size = XSTR_SPLIT_GROW_SIZE;
	array = xmalloc (sizeof(char*) * size);

	dup = xstr_dup (str);
	tok = strtok (dup, delim);
	array[0] = xstr_dup (tok);

	while (tok != NULL) {

		tok = strtok (NULL, delim);
		array[n] = xstr_dup (tok);
		n++;

		if (n == size) {
			if (OVERFLOW(XSTR_SPLIT_GROW_SIZE, size)) {
				warnx ("** xstr_split: integer overflow");
				exit (EXIT_FAILURE);
			}
			size = size + XSTR_SPLIT_GROW_SIZE;
			array = xrealloc (array, sizeof(char*) * size);
		}

	}

	free (dup);

	return array;
}


/* Free an array of strings.  */

void
xstr_freev (char **array)
{
	char **s;

	if (array == NULL) {
		return;
	}

	s = array;
	while (*s != NULL) {
		xfree (*s);
		s++;
	}
	xfree (array);
}


/* Return a new string based on a format string and arguments.  */

char*
xstr_dup_printf (const char *fmt, ...)
{
#ifdef __USE_GNU

	va_list args;
	char *s;

	va_start (args, fmt);

	if (vasprintf (&s, fmt, args) < 0) {
		warnx ("xstr_dup_printf failed");
		exit (EXIT_FAILURE);
	}

	va_end (args);

	return s;

#else

	/* (From the example in the Debian 3.1 vsprintf man page) */

	int n, size = 100; /* guess we need no more than 100 bytes */
	va_list args;
	char *s;

	s = xmalloc (size);
	for (;;) {

		va_start(args, fmt);
		n = vsnprintf (s, size, fmt, args);
		va_end(args);

		/* if that worked, return the string */
		if (n > -1 && n < size) {
			return s;
		}

		/* try again with more space */
		if (n > -1) {
			size = n + 1;
		}
		else {
			size = size * 2;
		}

		s = xrealloc (s, size);
	}
#endif
}


#ifdef TEST

static void
print_strings (char **str)
{
	while (*str != NULL) {
		printf ("[%s]\n", *str);
		str++;
	}
}


int
main (int argc, char *argv[])
{
	char *s, *t, **split;

	s = xstr_dup ("Hello, world!");
	printf ("[%s]\n", s);
	xfree (s);

	s = xstr_dup_printf ("%s %d", "the answer is usually", 42);
	printf ("%s\n", s);
	xfree (s);

	s = xstr_nfill (80, '-');
	printf ("%s\n", s);
	xfree (s);

	s = xstr_concat ("one", ", ", "two", ", ", "three", NULL);
	printf ("[%s]\n", s);
	xfree (s);

	s = xstr_join ("/", "foo", "bar", "baz", NULL);
	printf ("[%s]\n", s);
	xfree (s);

	split = xstr_split ("123, 432, 312, 654, 765, 223, 767", ", \t\n");
	print_strings (split);
	xstr_freev (split);

	split = xstr_split ("a b c d e", " ");
	s = xstr_joinv ("---", split);
	printf ("%s\n", s);
	xfree (s);
	xstr_freev (split);

	s = xstr_replace ("xaaxarxaz", "x", "f", 1);
	t = xstr_replace (s, "aa", "oo", 1);
	xfree (s);
	s = xstr_replace (t, "x", ", b", 0);
	xfree (t);
	printf ("%s\n", s);
	xfree (s);

	return EXIT_SUCCESS;
}
#endif

